import logging
from pathlib import Path
from typing import TYPE_CHECKING

from magnus import defaults
from magnus.catalog import BaseCatalog
from magnus.datastore import BaseRunLogStore
from magnus.secrets import BaseSecrets

if TYPE_CHECKING:
    from magnus.executor import BaseExecutor

from stevedore import extension

logger = logging.getLogger(defaults.NAME)


class BaseIntegration:
    """
    Base class for handling integration between Executor and one of Catalog, Secrets, RunLogStore.
    """
    mode_type = ''
    service_type = ''  # One of secret, catalog, datastore
    service_provider = ''  # The actual implementation of the service

    def __init__(self, executor: 'BaseExecutor', integration_service: object):
        self.executor = executor
        self.service = integration_service

    def validate(self, **kwargs):
        """
        Raise an exception if the mode_type is not compatible with service provider.

        By default, it is considered as compatible.
        """

    def configure_for_traversal(self, **kwargs):
        """
        Do any changes needed to both executor and service provider during traversal of the graph.

        By default, no change is required.
        """

    def configure_for_execution(self, **kwargs):
        """
        Do any changes needed to both executor and service provider during execution of a node.

        By default, no change is required.
        """


def get_service_type(service_provider: object) -> str:
    """
    Given a service provider, identify the type of service.

    Args:
        service_provider (object): The service provider object

    Raises:
        Exception: If the service provider is not inherited from one of BaseSecret, BaseCatalog, BaseRunLogStore

    Returns:
        [str]: Returns either 'secret', 'catalog', 'run_log_store' according to the service provider.
    """
    if isinstance(service_provider, BaseSecrets):
        return 'secrets'
    if isinstance(service_provider, BaseCatalog):
        return 'catalog'
    if isinstance(service_provider, BaseRunLogStore):
        return 'run_log_store'

    raise Exception('Service Provider is not a inherited from any of the Base Service providers')


def get_integration_handler(executor: 'BaseExecutor', service: object) -> BaseIntegration:
    """
    Return the integration handler between executor and the service.

    If none found to be implemented, return the BaseIntegration which does nothing.

    Args:
        executor (BaseExecutor): The executor
        service (object): The service provider

    Returns:
        [BaseIntegration]: The implemented integration handler or BaseIntegration if none found

    Raises:
        Exception: If multiple integrations are found for the executor and service
    """
    service_type = get_service_type(service)
    service_name = getattr(service, 'service_name')
    integrations = []

    mgr = extension.ExtensionManager(
        namespace="magnus.integration.BaseIntegration",
        invoke_on_load=True,
        invoke_kwds={'executor': executor, 'integration_service': service}
    )
    for name, kls in mgr.items():
        if (kls.obj.service_type == service_type and  # type: ignore
                kls.obj.mode_type == executor.service_name and  # type: ignore
                kls.obj.service_provider == service_name):
            logger.info(f'Identified an intergration pattern {kls.obj}')
            integrations.append(kls.obj)

    if len(integrations) > 1:
        msg = (
            f'Multiple integrations between {executor.service_name} and {service_name} of type {service_type} found. '
            'This is not correct. Please raise a bug report to fix this.'
        )
        logger.exception(msg)
        raise Exception(msg)

    if not integrations:
        logger.warning(
            f'Could not find an integration pattern for {executor.service_name} and {service_name} for {service_type}')  # type: ignore
        return BaseIntegration(executor, service)

    return integrations[0]


def validate(executor: 'BaseExecutor', service: object, **kwargs):
    """
    Helper function to resolve the Integration class and validate the compatibility between executor and service

    Args:
        executor (BaseExecutor) : The executor
        service (object): The service provider
    """
    integration_handler = get_integration_handler(executor, service)
    integration_handler.validate(**kwargs)


def configure_for_traversal(executor: 'BaseExecutor', service: object, **kwargs):
    """
    Helper function to resolve the Integration class and configure the executor and service for graph traversal

    Args:
        executor (BaseExecutor) : The executor
        service (object): The service provider
    """
    integration_handler = get_integration_handler(executor, service)
    integration_handler.configure_for_traversal(**kwargs)


def configure_for_execution(executor: 'BaseExecutor', service: object, **kwargs):
    """
    Helper function to resolve the Integration class and configure the executor and service for execution

    Args:
        executor (BaseExecutor) : The executor
        service (object): The service provider
    """
    integration_handler = get_integration_handler(executor, service)
    integration_handler.configure_for_execution(**kwargs)


class LocalComputeBufferedRunLogStore(BaseIntegration):
    """
    Local compute and buffered
    """
    mode_type = 'local'
    service_type = 'run_log_store'  # One of secret, catalog, datastore
    service_provider = 'buffered'  # The actual implementation of the service

    def validate(self, **kwargs):
        msg = (
            'Run log generated by buffered run log store are not persisted. '
            'Re-running this run, in case of a failure, is not possible'
        )
        logger.warning(msg)


class LocalComputeFileSystemRunLogStore(BaseIntegration):
    """
    Local compute and File system run log store
    """
    mode_type = 'local'
    service_type = 'run_log_store'  # One of secret, catalog, datastore
    service_provider = 'file-system'  # The actual implementation of the service

    def validate(self, **kwargs):
        if self.executor.is_parallel_execution():
            msg = (
                'Run log generated by file-system run log store are not thread safe. '
                'Inconsistent results are possible because of race conditions to write to the same file.\n'
                'Consider using partitioned run log store like database for consistent results.'

            )
            logger.warning(msg)


class LocalContainerComputeBufferedRunLogStore(BaseIntegration):
    """
    Only local execution mode is possible for Buffered Run Log store
    """
    mode_type = 'local-container'
    service_type = 'run_log_store'  # One of secret, catalog, datastore
    service_provider = 'buffered'  # The actual implementation of the service

    def validate(self, **kwargs):
        raise Exception('Only local compute mode is possible for buffered run log store')


class LocalContainerComputeFileSystemRunLogstore(BaseIntegration):
    """
    Integration between local container and file system run log store
    """
    mode_type = 'local-container'
    service_type = 'run_log_store'  # One of secret, catalog, datastore
    service_provider = 'file-system'  # The actual implementation of the service

    def validate(self, **kwargs):
        if self.executor.is_parallel_execution():
            msg = (
                'Run log generated by file-system run log store are not thread safe. '
                'Inconsistent results are possible because of race conditions to write to the same file.\n'
                'Consider using partitioned run log store like database for consistent results.'

            )
            logger.warning(msg)

    def configure_for_traversal(self, **kwargs):
        write_to = self.service.log_folder_name
        self.executor.volumes[str(Path(write_to).resolve())] = {
            'bind': f'{self.executor.container_log_location}',
            'mode': 'rw'
        }

    def configure_for_execution(self, **kwargs):
        self.service.config[self.service.CONFIG_KEY_LOG_FOLDER] = self.executor.container_log_location


class LocalContainerComputeDotEnvSecrets(BaseIntegration):
    """
    Integration between local container and dot env secrets
    """
    mode_type = 'local-container'
    service_type = 'secrets'  # One of secret, catalog, datastore
    service_provider = 'dotenv'  # The actual implementation of the service

    def validate(self, **kwargs):
        logger.warning('Using dot env for non local deployments is not ideal, consider options')

    def configure_for_traversal(self, **kwargs):
        secrets_location = self.service.get_secrets_location()
        self.executor.volumes[str(Path(secrets_location).resolve())] = {
            'bind': f'{self.executor.container_secrets_location}',
            'mode': 'ro'
        }

    def configure_for_execution(self, **kwargs):
        self.service.config['location'] = self.executor.container_secrets_location


class LocalContainerDoNothingCatalog(BaseIntegration):
    """
    Integration between local container and do nothing catalog
    """
    mode_type = 'local-container'
    service_type = 'catalog'  # One of secret, catalog, datastore
    service_provider = 'do-nothing'  # The actual implementation of the service

    def validate(self, **kwargs):
        msg = (
            'A do-nothing catalog does not hold any data and therefore cannot pass data between nodes.'
        )
        logger.warning(msg)


class LocalDoNothingCatalog(BaseIntegration):
    """
    Integration between local and do nothing catalog
    """
    mode_type = 'local'
    service_type = 'catalog'  # One of secret, catalog, datastore
    service_provider = 'do-nothing'  # The actual implementation of the service

    def validate(self, **kwargs):
        msg = (
            'A do-nothing catalog does not hold any data and therefore cannot pass data between nodes.'
        )
        logger.warning(msg)


class LocalContainerComputeFileSystemCatalog(BaseIntegration):
    """
    Integration pattern between Local container and File System catalog
    """
    mode_type = 'local-container'
    service_type = 'catalog'  # One of secret, catalog, datastore
    service_provider = 'file-system'  # The actual implementation of the service

    def configure_for_traversal(self, **kwargs):
        catalog_location = self.service.get_catalog_location()
        self.executor.volumes[str(Path(catalog_location).resolve())] = {
            'bind': f'{self.executor.container_catalog_location}',
            'mode': 'rw'
        }

    def configure_for_execution(self, **kwargs):
        self.service.config['catalog_location'] = self.executor.container_catalog_location


class DemoRenderBufferedRunLogStore(BaseIntegration):
    """
    Demo rendered and buffered
    """
    mode_type = 'demo-renderer'
    service_type = 'run_log_store'  # One of secret, catalog, datastore
    service_provider = 'buffered'  # The actual implementation of the service

    def validate(self, **kwargs):
        msg = (
            'Run log generated by buffered run log store are not persisted. '
            'Demo renderer cannot use buffered as steps are executed as individual commands.'
        )
        logger.exception(msg)
        raise Exception(msg)



import logging
from logging.config import fileConfig

import click
from click_plugins import with_plugins
from pkg_resources import iter_entry_points, resource_filename

from magnus import defaults, pipeline

fileConfig(resource_filename(__name__, 'log_config.ini'))
logger = logging.getLogger(defaults.NAME)



@with_plugins(iter_entry_points('magnus.cli_plugins'))
@click.group()
@click.version_option()
def cli():
    """
    Welcome to magnus. Please provide the command that you want to use.
    All commands have options that you can see by magnus <command> --help
    """
    pass


@cli.command('execute', short_help="Execute/translate a pipeline")
@click.option('-f', '--file', default='pipeline.yaml', help='The pipeline definition file', show_default=True)
@click.option('-v', '--var-file', default=None, help='Variables used in pipeline definition', show_default=True)
@click.option('-c', '--config-file', default=None,
              help="config file, in yaml, to be used for the run", show_default=True)
@click.option('-p', '--parameters-file', default=None,
              help="Parameters, in yaml,  accessible by the application", show_default=True)
@click.option('--log-level', default=defaults.LOG_LEVEL, help='The log level', show_default=True,
              type=click.Choice(['INFO', 'DEBUG', 'WARNING', 'ERROR', 'FATAL']))
@click.option('--tag', help='A tag attached to the run')
@click.option('--run-id', help='An optional run_id, one would be generated if not provided')
@click.option('--use-cached', help='Provide the previous run_id to re-run.', show_default=True)
def execute(file, var_file, config_file, parameters_file, log_level, tag, run_id, use_cached):  # pragma: no cover
    """
    Entry point to executing a pipeline. This command is most commonly used
    either to execute a pipeline or to translate the pipeline definition to another language.
    """
    logger.setLevel(log_level)
    pipeline.execute(
        variables_file=var_file, configuration_file=config_file, pipeline_file=file, tag=tag,
        run_id=run_id, use_cached=use_cached, parameters_file=parameters_file)


@cli.command("execute_single_node", short_help="Internal entry point to execute a single node", hidden=True)
@click.argument('run_id')
@click.argument('step_name')
@click.option('--map-variable', default='', help='The map variable dictionary in str', show_default=True)
@click.option('-f', '--file', default='pipeline.yaml', help='The pipeline definition file', show_default=True)
@click.option('-v', '--var-file', default=None, help='Variables used in pipeline definition', show_default=True)
@click.option('-c', '--config-file', default=None,
              help="config file, in yaml, to be used for the run", show_default=True)
@click.option('-p', '--parameters-file', default=None,
              help="Parameters, in yaml,  accessible by the application", show_default=True)
@click.option('--log-level', default=defaults.LOG_LEVEL, help='The log level', show_default=True,
              type=click.Choice(['INFO', 'DEBUG', 'WARNING', 'ERROR', 'FATAL']))
@click.option('--tag', default='', help='A tag attached to the run')
def execute_single_node(run_id, step_name, map_variable, file, var_file, config_file, parameters_file, log_level, tag):
    logger.setLevel(log_level)

    pipeline.execute_single_node(variables_file=var_file,
                                 configuration_file=config_file, pipeline_file=file,
                                 step_name=step_name, map_variable=map_variable,
                                 run_id=run_id, tag=tag, parameters_file=parameters_file)


@cli.command("execute_single_branch", short_help="Internal entry point to execute a single branch", hidden=True)
@click.argument('run_id')
@click.argument('branch_name')
@click.option('--map-variable', default='', help='The map variable dictionary in str', show_default=True)
@click.option('-f', '--file', default='pipeline.yaml', help='The pipeline definition file', show_default=True)
@click.option('-v', '--var-file', default=None, help='Variables used in pipeline definition', show_default=True)
@click.option('-c', '--config-file', default=None,
              help="config file, in yaml, to be used for the run", show_default=True)
@click.option('--log-level', default=defaults.LOG_LEVEL, help='The log level', show_default=True,
              type=click.Choice(['INFO', 'DEBUG', 'WARNING', 'ERROR', 'FATAL']))
def execute_single_branch(run_id, branch_name, map_variable, file, var_file, config_file, log_level):
    logger.setLevel(log_level)

    pipeline.execute_single_brach(variables_file=var_file, configuration_file=config_file,
                                  pipeline_file=file, branch_name=branch_name, map_variable=map_variable,
                                  run_id=run_id)


# @cli.command("build_docker_image", short_help="Utility tool to build docker images")
def build_docker_image():
    """
    Utility tool to build a docker image.

    The design thought is to use pyproject.toml to build the docker image.
    """
    pass




import logging

from yachalk import chalk

from magnus.interaction import (get_from_catalog, get_parameter, get_secret,
                                put_in_catalog, store_parameter, track_this)

chalk_colors = {
    'debug': chalk.grey,
    'info': chalk.green,
    'warning': chalk.yellow_bright,
    'error': chalk.bold.red
}


class ColorFormatter(logging.Formatter):
    """
    Custom class to get colors to logs
    """

    def __init__(self, *args, **kwargs):
        # can't do super(...) here because Formatter is an old school class
        logging.Formatter.__init__(self, *args, **kwargs)

    def format(self, record):
        levelname = record.levelname
        color = chalk_colors[levelname.lower()]
        message = logging.Formatter.format(self, record)
        return color(message)


logging.ColorFormatter = ColorFormatter  # type: ignore



import importlib
import json
import logging
import os
import subprocess
import sys

from magnus import defaults, utils

logger = logging.getLogger(defaults.NAME)

try:
    import papermill as pm
except ImportError:
    pm = None


class BaseTaskType:  # pylint: disable=too-few-public-methods
    """
    A base task class which does the execution of command defined by the user
    """
    task_type = ''

    def __init__(self, command: str, config: dict = None):
        self.command = command
        self.config = config or {}

    def get_parameters(self, map_variable: dict = None, **kwargs) -> dict:
        """
        Return the parameters in scope for the execution

        Args:
            map_variable (dict, optional): If the command is part of map node, the value of map. Defaults to None.

        Returns:
            dict: The parameters dictionary in-scope for the task execution
        """
        return utils.get_user_set_parameters(remove=False)

    def execute_command(self, map_variable: dict = None, **kwargs):
        """
        The function to execute the command.

        And map_variable is sent in as an argument into the function.

        Args:
            map_variable (dict, optional): If the command is part of map node, the value of map. Defaults to None.

        Raises:
            NotImplementedError: Base class, not implemented
        """
        raise NotImplementedError()

    def set_parameters(self, parameters: dict = None, **kwargs):
        """
        Set the parameters back to the environment variables.

        Args:
            parameters (dict, optional): The parameters to set back as env variables. Defaults to None.
        """
        # Nothing to do
        if not parameters:
            return

        if not isinstance(parameters, dict):
            msg = (
                f'call to function {self.command} returns of type: {type(parameters)}. '
                'Only dictionaries are supported as return values for functions as part part of magnus pipeline.')
            raise Exception(msg)

        for key, value in parameters.items():
            logger.info(f'Setting User defined parameter {key} with value: {value}')
            os.environ[defaults.PARAMETER_PREFIX + key] = json.dumps(value)


class PythonTaskType(BaseTaskType):  # pylint: disable=too-few-public-methods
    """
    The task class for python command
    """
    task_type = 'python'

    def execute_command(self, map_variable: dict = None, **kwargs):
        module, func = utils.get_module_and_func_names(self.command)
        sys.path.insert(0, os.getcwd())  # Need to add the current directory to path
        imported_module = importlib.import_module(module)
        f = getattr(imported_module, func)

        parameters = self.get_parameters()
        filtered_parameters = utils.filter_arguments_for_func(f, parameters, map_variable)

        logger.info(f'Calling {func} from {module} with {filtered_parameters}')
        try:
            user_set_parameters = f(**filtered_parameters)
        except Exception as _e:
            msg = (
                f'Call to the function {self.command} with {filtered_parameters} did not succeed.\n'
            )
            logger.exception(msg)
            logger.exception(_e)
            raise

        self.set_parameters(user_set_parameters)


class PythonLambdaTaskType(BaseTaskType):  # pylint: disable=too-few-public-methods
    """
    The task class for python-lambda command
    """
    task_type = 'python-lambda'

    def execute_command(self, map_variable: dict = None, **kwargs):
        if '_' in self.command or '__' in self.command:
            msg = (
                f'Command given to {self.task_type} cannot have _ or __ in them. '
                'The string is supposed to be for simple expressions only.'
            )
            raise Exception(msg)

        f = eval(self.command)

        parameters = self.get_parameters()
        filtered_parameters = utils.filter_arguments_for_func(f, parameters, map_variable)

        logger.info(f'Calling lambda function: {self.command} with {filtered_parameters}')
        try:
            user_set_parameters = f(**filtered_parameters)
        except Exception as _e:
            msg = (
                f'Call to the function {self.command} with {filtered_parameters} did not succeed.\n'
            )
            logger.exception(msg)
            logger.exception(_e)
            raise

        self.set_parameters(user_set_parameters)


class NotebookTaskType(BaseTaskType):
    """
    The task class for Notebook based execution
    """
    task_type = 'notebook'

    def __init__(self, command: str, config: dict = None):
        if not command.endswith('.ipynb'):
            raise Exception('Notebook task should point to a ipynb file')

        super().__init__(command, config)

    def execute_command(self, map_variable: dict = None, **kwargs):
        try:
            if not pm:
                raise ImportError('Papermill is required for notebook type node')

            parameters = self.get_parameters()

            notebook_parameters = pm.inspect_notebook(self.command)
            filtered_parameters = utils.filter_arguments_from_parameters(parameters=parameters,
                                                                         signature_parameters=notebook_parameters,
                                                                         map_variable=map_variable)
            notebook_output_path = self.config.get('notebook_output_path',
                                                   ''.join(self.command.split('.')[:-1]) + '_out.ipynb')
            kernel = self.config.get('notebook_kernel', None)
            papermill_optional_args = self.config.get('optional_papermill_args', {})

            kwds = {
                'input_path': self.command,
                'output_path': notebook_output_path,
                'parameters': filtered_parameters,
            }

            kwds.update(papermill_optional_args)

            if kernel:
                kwds['kernel_name'] = kernel

            pm.execute_notebook(**kwds)
        except ImportError as e:
            msg = (
                f'Task type of notebook requires papermill to be installed. Please install via optional: notebook'
            )
            raise Exception(msg) from e


class ShellTaskType(BaseTaskType):
    """
    The task class for shell based commands
    """
    task_type = 'shell'

    def execute_command(self, map_variable: dict = None, **kwargs):
        # TODO can we do this without shell=True. Hate that but could not find a way out
        # This is horribly weird, focussing only on python ways of doing for now
        # It might be that we have to write a bash/windows script that does things for us
        # Need to over-ride set parameters too
        subprocess_env = os.environ.copy()
        if map_variable:
            subprocess_env[defaults.PARAMETER_PREFIX + 'MAP_VARIABLE'] = json.dumps(map_variable)
        result = subprocess.run(self.command, check=True, env=subprocess_env, shell=True,
                                stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
        print(result.stdout)
        print(result.stderr)



import importlib
import json
import logging
import multiprocessing
import os
import shlex
import subprocess
import sys
from collections import OrderedDict
from datetime import datetime
from typing import List, Type, Union

from pkg_resources import resource_filename

import magnus
from magnus import defaults, utils
from magnus.graph import create_graph

logger = logging.getLogger(defaults.NAME)


class BaseNode:
    """
    Base class with common functionality provided for a Node of a graph.

    A node of a graph could be a
        * single execution node as task, success, fail.
        * Could be graph in itself as parallel, dag and map.
        * could be a convenience function like as-is.

    The name is relative to the DAG.
    The internal name of the node, is absolute name in dot path convention.
        This has one to one mapping to the name in the run log
    The internal name of a node, should always be odd when split against dot.

    The internal branch name, only applies for branched nodes, is the branch it belongs to.
    The internal branch name should always be even when split against dot.
    """

    node_type = ''

    def __init__(self, name, internal_name, config, execution_type, internal_branch_name=None):
        # pylint: disable=R0914,R0913
        self.name = name
        self.internal_name = internal_name  #  Dot notation naming of the steps
        self.config = config
        self.internal_branch_name = internal_branch_name  # parallel, map, dag only have internal names
        self.execution_type = execution_type
        self.branches = None
        self.is_composite = False

    def command_friendly_name(self, replace_with=defaults.COMMAND_FRIENDLY_CHARACTER) -> str:
        """
        Replace spaces with special character for spaces.
        Spaces in the naming of the node is convenient for the user but causes issues when used programmatically.

        Returns:
            str: The command friendly name of the node
        """
        return self.internal_name.replace(' ', replace_with)

    @ classmethod
    def get_internal_name_from_command_name(cls, command_name: str) -> str:
        """
        Replace Magnus specific character (%) with whitespace.
        The opposite of command_friendly_name.

        Args:
            command_name (str): The command friendly node name

        Returns:
            str: The internal name of the step
        """
        return command_name.replace(defaults.COMMAND_FRIENDLY_CHARACTER, ' ')

    @ classmethod
    def resolve_map_placeholders(cls, name: str, map_variable: dict = None) -> str:
        """
        If there is no map step used, then we just return the name as we find it.

        If there is a map variable being used, replace every occurrence of the map variable placeholder with
        the value sequentially.

        For example:
        1). dag:
              start_at: step1
              steps:
                step1:
                    type: map
                    iterate_on: y
                    iterate_as: y_i
                    branch:
                        start_at: map_step1
                        steps:
                            map_step1: # internal_name step1.placeholder.map_step1
                                type: task
                                command: a.map_func
                                command_type: python
                                next: map_success
                            map_success:
                                type: success
                            map_failure:
                                type: fail

            and if y is ['a', 'b', 'c'].

            This method would be called 3 times with map_variable = {'y_i': 'a'}, map_variable = {'y_i': 'b'} and
            map_variable = {'y_i': 'c'} corresponding to the three branches.

        For nested map branches, we would get the map_variables ordered heirarchically.

        Args:
            name (str): The name to resolve
            map_variable (dict): The dictionary of map variables

        Returns:
            [str]: The resolved name
        """
        if not map_variable:
            return name

        for _, value in map_variable.items():
            name = name.replace(defaults.MAP_PLACEHOLDER, value, 1)

        return name

    def get_step_log_name(self, map_variable: dict = None) -> str:
        """
        For every step in the dag, there is a corresponding step log name.
        This method returns the step log name in dot path convention.

        All node types except a map state has a "static" defined step_log names and are equivalent to internal_name.
        For nodes belonging to map state, the internal name has a placeholder that is replaced at runtime.

        Args:
            map_variable (dict): If the node is of type map, the names are based on the current iteration state of the
            parameter.

        Returns:
            str: The dot path name of the step log name
        """
        return self.resolve_map_placeholders(self.internal_name, map_variable=map_variable)

    def get_branch_log_name(self, map_variable: dict = None) -> str:
        """
        For nodes that are internally branches, this method returns the branch log name.
        The branch log name is in dot path convention.

        For nodes that are not map, the internal branch name is equivalent to the branch name.
        For map nodes, the internal branch name has a placeholder that is replaced at runtime.

        Args:
            map_variable (dict): If the node is of type map, the names are based on the current iteration state of the
            parameter.

        Returns:
            str: The dot path name of the branch log
        """
        return self.resolve_map_placeholders(self.internal_branch_name, map_variable=map_variable)

    def __str__(self):  # pragma: no cover
        return f'Node of type {self.node_type} and name {self.internal_name}'

    def get_on_failure_node(self) -> Union[str, None]:
        """
        If the node defines a on_failure node in the config, return this or None.

        The naming is relative to the dag, the caller is supposed to resolve it to the correct graph

        Returns:
            str: The on_failure node defined by the dag or ''
        """
        if 'on_failure' in self.config:
            return self.config['on_failure']
        return None

    def get_catalog_settings(self) -> Union[dict, None]:
        """
        If the node defines a catalog settings, return it or None

        Returns:
            dict: catalog settings defined as per the node or None
        """
        if 'catalog' in self.config:
            return self.config['catalog']
        return None

    def get_branch_by_name(self, branch_name: str):
        """
        Retrieve a branch by name.

        The name is expected to follow a dot path convention.

        This method will raise an exception if the node does not have any branches.
        i.e: task, success, fail and as-is would raise an exception

        Args:
            branch_name (str): [description]

        Raises:
            Exception: [description]
        """
        raise Exception(f'Node of type {self.node_type} does not have any branches')

    def is_terminal_node(self):
        """
        Returns whether a node has a next node

        Returns:
            bool: True or False of whether there is next node.
        """
        if 'next' in self.config:
            return False
        return True

    def get_neighbors(self):
        """
        Gets the connecting neighbor nodes, either the "next" node or "on_failure" node.

        Returns:
            list: List of connected neighbors for a given node. Empty if terminal node.
        """
        neighbors = []
        next_node = self.get_next_node()
        if next_node:
            neighbors += [next_node]

        fail_node = self.get_on_failure_node()
        if fail_node:
            neighbors += [fail_node]

        return neighbors

    def get_next_node(self) -> Union[str, None]:
        """
        Return the next node as defined by the config.

        Returns:
            str: The node name, relative to the dag, as defined by the config
        """
        if not self.is_terminal_node():
            return self.config['next']
        return None

    def get_mode_config(self, mode_type) -> dict:
        """
        Return the mode config of the node, if defined, or empty dict

        Args:
            mode_type (str): The mode type that the config refers to.

        Returns:
            dict: The mode config, if defined or an empty dict
        """
        if 'mode_config' in self.config:
            return self.config['mode_config'].get(mode_type, {}) or {}
        return {}

    def get_max_attempts(self) -> int:
        """
        The number of max attempts as defined by the config or 1.

        Returns:
            int: The number of maximum retries as defined by the config or 1.
        """
        if 'retry' in self.config:
            return int(self.config['retry']) or 1
        return 1

    def execute(self, executor, mock=False, map_variable: dict = None, **kwargs):
        """
        The actual function that does the execution of the command in the config.

        Should only be implemented for task, success, fail and as-is and never for
        composite nodes.

        Args:
            executor (magnus.executor.BaseExecutor): The executor mode class
            mock (bool, optional): Don't run, just pretend. Defaults to False.
            map_variable (str, optional): The value of the map iteration variable, if part of a map node.
                Defaults to ''.

        Raises:
            NotImplementedError: Base class, hence not implemented.
        """
        raise NotImplementedError

    def execute_as_graph(self, executor, map_variable: dict = None, **kwargs):
        """
        This function would be called to set up the execution of the individual
        branches of a composite node.

        Function should only be implemented for composite nodes like dag, map, parallel.

        Args:
            executor (magnus.executor.BaseExecutor): The executor mode.

        Raises:
            NotImplementedError: Base class, hence not implemented.
        """
        raise NotImplementedError


def validate_node(node: BaseNode) -> List[str]:
    """
    Given a node defintion, run it against a specification of fields that are
    required and should not be present.

    Args:
        node (BaseNode): The node object created before validation

    Raises:
        Exception: If the node type is not part of the specs

    Returns:
        List[str]: The list of error messages, if found
    """
    specs = utils.load_yaml(resource_filename(__name__, defaults.NODE_SPEC_FILE))
    if node.node_type not in specs:
        raise Exception('Undefined node type, please update specs')

    node_spec = specs[node.node_type]
    messages = []
    if '.' in node.name:
        messages.append('Node names cannot have . in them')

    if '%' in node.name:
        messages.append("Node names cannot have '%' in them")

    if 'required' in node_spec:
        for req in node_spec['required']:
            if not req in node.config:
                messages.append(f'{node.name} should have {req} field')
                continue

    if 'error_on' in node_spec:
        for err in node_spec['error_on']:
            if err in node.config:
                messages.append(f'{node.name} should not have {err} field')
    return messages


class TaskNode(BaseNode):
    """
    A node of type Task.

    This node does the actual function execution of the graph in all cases.
    """
    node_type = 'task'

    def execute(self, executor, mock=False, map_variable: dict = None, **kwargs):
        # Here is where the juice is
        attempt_log = executor.run_log_store.create_attempt_log()
        try:
            attempt_log.start_time = str(datetime.now())
            attempt_log.status = defaults.SUCCESS
            if not mock:
                # Do not run if we are mocking the execution, could be useful for caching and dry runs
                self.execution_type.execute_command(step_config=self.config, map_variable=map_variable)
        except Exception as _e:  # pylint: disable=W0703
            logger.exception('Task failed')
            attempt_log.status = defaults.FAIL
            attempt_log.message = str(_e)
        finally:
            attempt_log.end_time = str(datetime.now())
            attempt_log.duration = utils.get_duration_between_datetime_strings(
                attempt_log.start_time, attempt_log.end_time)
        return attempt_log

    def execute_as_graph(self, executor, map_variable: dict = None, **kwargs):
        """
        Should not be implemented for a single node.

        Args:
            executor ([type]): [description]

        Raises:
            Exception: Not a composite node, always raises an exception
        """
        raise Exception('Node is not a composite node, invalid traversal rule')


class FailNode(BaseNode):
    """
    A leaf node of the graph that represents a failure node
    """
    node_type = 'fail'

    def execute(self, executor, mock=False, map_variable: dict = None, **kwargs):
        attempt_log = executor.run_log_store.create_attempt_log()
        try:
            attempt_log.start_time = str(datetime.now())
            attempt_log.status = defaults.SUCCESS
            #  could be a branch or run log
            run_or_branch_log = executor.run_log_store.get_branch_log(
                self.get_branch_log_name(map_variable), executor.run_id)
            run_or_branch_log.status = defaults.FAIL
            executor.run_log_store.add_branch_log(run_or_branch_log, executor.run_id)
        except BaseException:  # pylint: disable=W0703
            logger.exception('Fail node execution failed')
        finally:
            attempt_log.status = defaults.SUCCESS  # This is a dummy node, so we ignore errors and mark SUCCESS
            attempt_log.end_time = str(datetime.now())
            attempt_log.duration = utils.get_duration_between_datetime_strings(
                attempt_log.start_time, attempt_log.end_time)
        return attempt_log

    def execute_as_graph(self, executor, map_variable: dict = None, **kwargs):
        """
        Should not be implemented for a single node.

        Args:
            executor ([type]): [description]

        Raises:
            Exception: Not a composite node, always raises an exception
        """
        raise Exception('Node is not a composite node, invalid traversal rule')


class SuccessNode(BaseNode):
    """
    A leaf node of the graph that represents a success node
    """
    node_type = 'success'

    def execute(self, executor, mock=False, map_variable: dict = None, **kwargs):
        attempt_log = executor.run_log_store.create_attempt_log()
        try:
            attempt_log.start_time = str(datetime.now())
            attempt_log.status = defaults.SUCCESS
            #  could be a branch or run log
            run_or_branch_log = executor.run_log_store.get_branch_log(
                self.get_branch_log_name(map_variable), executor.run_id)
            run_or_branch_log.status = defaults.SUCCESS
            executor.run_log_store.add_branch_log(run_or_branch_log, executor.run_id)
        except BaseException:  # pylint: disable=W0703
            logger.exception('Success node execution failed')
        finally:
            attempt_log.status = defaults.SUCCESS  # This is a dummy node and we make sure we mark it as success
            attempt_log.end_time = str(datetime.now())
            attempt_log.duration = utils.get_duration_between_datetime_strings(
                attempt_log.start_time, attempt_log.end_time)
        return attempt_log

    def execute_as_graph(self, executor, map_variable: dict = None, **kwargs):
        """
        Should not be implemented for a single node.

        Args:
            executor ([type]): [description]

        Raises:
            Exception: Not a composite node, always raises an exception
        """
        raise Exception('Node is not a composite node, invalid traversal rule')


class ParallelNode(BaseNode):
    """
    A composite node containing many graph objects within itself.

    The structure is generally:
        ParallelNode:
            Branch A:
                Sub graph definition
            Branch B:
                Sub graph definition
            . . .

    We currently support parallel nodes within parallel nodes.
    """
    node_type = 'parallel'

    def __init__(self, name, internal_name, config, execution_type, internal_branch_name=None):
        # pylint: disable=R0914,R0913
        super().__init__(name, internal_name, config, execution_type, internal_branch_name=internal_branch_name)
        self.branches = self.get_sub_graphs()
        self.is_composite = True

    def get_sub_graphs(self):
        """
        For the branches mentioned in the config['branches'], create a graph object.
        The graph object is also given an internal naming convention following a dot path convention

        Returns:
            dict: A branch_name: dag for every branch mentioned in the branches
        """

        branches = {}
        for branch_name, branch_config in self.config['branches'].items():
            sub_graph = create_graph(branch_config, internal_branch_name=self.internal_name + '.' + branch_name)
            branches[self.internal_name + '.' + branch_name] = sub_graph

        if not branches:
            raise Exception('A parallel node should have branches')
        return branches

    def get_branch_by_name(self, branch_name: str):
        """
        Retrieve a branch by name.
        The name is expected to follow a dot path convention.

        Returns a Graph Object

        Args:
            branch_name (str): The name of the branch to retrieve

        Raises:
            Exception: If the branch by that name does not exist
        """
        if branch_name in self.branches:
            return self.branches[branch_name]

        raise Exception(f'No branch by name: {branch_name} is present in {self.name}')

    def execute(self, executor, mock=False, map_variable: dict = None, **kwargs):
        """
        This method should never be called for a node of type Parallel

        Args:
            executor (BaseExecutor): The Executor class as defined by the config
            mock (bool, optional): If the operation is just a mock. Defaults to False.

        Raises:
            NotImplementedError: This method should never be called for a node of type Parallel
        """
        raise Exception('Node is of type composite, error in traversal rules')

    def execute_as_graph(self, executor, map_variable: dict = None, **kwargs):
        """
        This function does the actual execution of the sub-branches of the parallel node.

        From a design perspective, this function should not be called if the execution mode is 3rd party orchestrated.

        The modes that render the job specifications, do not need to interact with this node at all as they have their
        own internal mechanisms of handing parallel states.
        If they do not, you can find a way using as-is nodes as hack nodes.

        The execution of a dag, could result in
            * The dag being completely executed with a definite (fail, success) state in case of
                local or local-container execution
            * The dag being in a processing state with PROCESSING status in case of local-aws-batch

        Only fail state is considered failure during this phase of execution.

        Args:
            executor (Executor): The Executor as per the use config
            **kwargs: Optional kwargs passed around
        """
        # Prepare the branch logs
        for internal_branch_name, branch in self.branches.items():
            effective_branch_name = self.resolve_map_placeholders(internal_branch_name, map_variable=map_variable)

            branch_log = executor.run_log_store.create_branch_log(effective_branch_name)
            branch_log.status = defaults.PROCESSING
            executor.run_log_store.add_branch_log(branch_log, executor.run_id)

        jobs = []
        # Given that we can have nesting and complex graphs, controlling the number of processess is hard.
        # A better way is to actually submit the job to some process scheduler which does resource management
        for internal_branch_name, branch in self.branches.items():
            if executor.is_parallel_execution():
                # Trigger parallel jobs
                action = magnus.pipeline.execute_single_brach
                kwargs = {
                    'configuration_file': executor.configuration_file,
                    'pipeline_file': executor.pipeline_file,
                    'variables_file': executor.variables_file,
                    'branch_name': internal_branch_name.replace(' ', defaults.COMMAND_FRIENDLY_CHARACTER),
                    'run_id': executor.run_id,
                    'map_variable': json.dumps(map_variable)
                }
                process = multiprocessing.Process(target=action, kwargs=kwargs)
                jobs.append(process)
                process.start()

            else:
                # If parallel is not enabled, execute them sequentially
                executor.execute_graph(branch, map_variable=map_variable, **kwargs)

        for job in jobs:
            job.join()  # Find status of the branches

        step_success_bool = True
        waiting = False
        for internal_branch_name, branch in self.branches.items():
            effective_branch_name = self.resolve_map_placeholders(internal_branch_name, map_variable=map_variable)
            branch_log = executor.run_log_store.get_branch_log(effective_branch_name, executor.run_id)
            if branch_log.status == defaults.FAIL:
                step_success_bool = False

            if branch_log.status == defaults.PROCESSING:
                waiting = True

        # Collate all the results and update the status of the step
        effective_internal_name = self.resolve_map_placeholders(self.internal_name, map_variable=map_variable)
        step_log = executor.run_log_store.get_step_log(effective_internal_name, executor.run_id)
        step_log.status = defaults.PROCESSING

        if step_success_bool:  #  If none failed and nothing is waiting
            if not waiting:
                step_log.status = defaults.SUCCESS
        else:
            step_log.status = defaults.FAIL

        executor.run_log_store.add_step_log(step_log, executor.run_id)


class MapNode(BaseNode):
    """
    A composite node that contains ONE graph object within itself that has to be executed with an iterable.

    The structure is genrally:
        MapNode:
            branch

        The config is expected to have a variable 'iterate_on' and iterate_as which are looked for in the parameters.
        for iter_variable in parameters['iterate_on']:
            Execute the Branch by sending {'iterate_as': iter_variable}

    The internal naming convention creates branches dynamically based on the iteration value
    """
    node_type = 'map'

    def __init__(self, name, internal_name, config, execution_type, internal_branch_name=None):
        # pylint: disable=R0914,R0913
        super().__init__(name, internal_name, config, execution_type, internal_branch_name=internal_branch_name)
        self.iterate_on = self.config.get('iterate_on', None)
        self.iterate_as = self.config.get('iterate_as', None)
        self.is_composite = True

        if not self.iterate_on:
            raise Exception('A node type of map requires a parameter iterate_on, please define it in the config')
        if not self.iterate_as:
            raise Exception('A node type of map requires a parameter iterate_as, please define it in the config')

        self.branch_placeholder_name = defaults.MAP_PLACEHOLDER
        self.branch = self.get_sub_graph()

    def get_sub_graph(self):
        """
        Create a sub-dag from the config['branch']

        The graph object has an internal branch name, that is equal to the name of the step.
        And the sub-dag nodes follow an dot path naming convention

        Returns:
            Graph: A graph object
        """

        branch_config = self.config['branch']
        branch = create_graph(
            branch_config, internal_branch_name=self.internal_name + '.' + self.branch_placeholder_name)
        return branch

    def get_branch_by_name(self, branch_name: str):
        """
        Retrieve a branch by name.

        In the case of a Map Object, the branch naming is dynamic as it is parameterized on iterable.
        This method takes no responsibility in checking the validity of the naming.

        Returns a Graph Object

        Args:
            branch_name (str): The name of the branch to retrieve

        Raises:
            Exception: If the branch by that name does not exist
        """
        return self.branch

    def execute(self, executor, mock=False, map_variable: dict = None, **kwargs):
        """
        This method should never be called for a node of type map

        Args:
            executor (BaseExecutor): The Executor class as defined by the config
            mock (bool, optional): If the operation is just a mock. Defaults to False.

        Raises:
            NotImplementedError: This method should never be called for a node of type Parallel
        """
        raise Exception('Node is of type composite, error in traversal rules')

    def execute_as_graph(self, executor, map_variable: dict = None, **kwargs):
        """
        This function does the actual execution of the branch of the map node.

        From a design perspective, this function should not be called if the execution mode is 3rd party orchestrated.

        The modes that render the job specifications, do not need to interact with this node at all as
        they have their own internal mechanisms of handing map states or dynamic parallel states.
        If they do not, you can find a way using as-is nodes as hack nodes.

        The actual logic is :
            * We iterate over the iterable as mentioned in the config
            * For every value in the iterable we call the executor.execute_graph(branch, iterate_as: iter_variable)

        The execution of a dag, could result in
            * The dag being completely executed with a definite (fail, success) state in case of local
                or local-container execution
            * The dag being in a processing state with PROCESSING status in case of local-aws-batch

        Only fail state is considered failure during this phase of execution.

        Args:
            executor (Executor): The Executor as per the use config
            map_variable (dict): The map variables the graph belongs to
            **kwargs: Optional kwargs passed around
        """
        run_log = executor.run_log_store.get_run_log_by_id(executor.run_id)
        if self.iterate_on not in run_log.parameters:
            raise Exception(
                f'Expected parameter {self.iterate_on} not present in Run Log parameters, was it ever set before?')

        iterate_on = run_log.parameters[self.iterate_on]
        if not isinstance(iterate_on, list):
            raise Exception('Only list is allowed as a valid iterator type')

        # Prepare the branch logs
        for iter_variable in iterate_on:
            effective_branch_name = self.resolve_map_placeholders(
                self.internal_name + '.' + str(iter_variable),
                map_variable=map_variable)
            branch_log = executor.run_log_store.create_branch_log(effective_branch_name)
            branch_log.status = defaults.PROCESSING
            executor.run_log_store.add_branch_log(branch_log, executor.run_id)

        jobs = []
        # Given that we can have nesting and complex graphs, controlling the number of processess is hard.
        # A better way is to actually submit the job to some process scheduler which does resource management
        for iter_variable in iterate_on:
            effective_map_variable = map_variable or OrderedDict()
            effective_map_variable[self.iterate_as] = iter_variable

            if executor.is_parallel_execution():
                # Trigger parallel jobs
                action = magnus.pipeline.execute_single_brach
                kwargs = {
                    'configuration_file': executor.configuration_file,
                    'pipeline_file': executor.pipeline_file,
                    'variables_file': executor.variables_file,
                    'branch_name': self.branch.internal_branch_name.replace(' ', defaults.COMMAND_FRIENDLY_CHARACTER),
                    'run_id': executor.run_id,
                    'map_variable': json.dumps(effective_map_variable)
                }
                process = multiprocessing.Process(target=action, kwargs=kwargs)
                jobs.append(process)
                process.start()

            else:
                # If parallel is not enabled, execute them sequentially
                executor.execute_graph(self.branch, map_variable=effective_map_variable, **kwargs)

        for job in jobs:
            job.join()
        # # Find status of the branches
        step_success_bool = True
        waiting = False
        for iter_variable in iterate_on:
            effective_branch_name = self.resolve_map_placeholders(self.internal_name + '.' + str(iter_variable),
                                                                  map_variable=map_variable)
            branch_log = executor.run_log_store.get_branch_log(
                effective_branch_name, executor.run_id)
            if branch_log.status == defaults.FAIL:
                step_success_bool = False

            if branch_log.status == defaults.PROCESSING:
                waiting = True

        # Collate all the results and update the status of the step
        effective_internal_name = self.resolve_map_placeholders(self.internal_name, map_variable=map_variable)
        step_log = executor.run_log_store.get_step_log(effective_internal_name, executor.run_id)
        step_log.status = defaults.PROCESSING

        if step_success_bool:  #  If none failed and nothing is waiting
            if not waiting:
                step_log.status = defaults.SUCCESS
        else:
            step_log.status = defaults.FAIL

        executor.run_log_store.add_step_log(step_log, executor.run_id)


class DagNode(BaseNode):
    """
    A composite node that internally holds a dag.

    The structure is genrally:
        DagNode:
            dag_definition: A YAML file that holds the dag in 'dag' block

        The config is expected to have a variable 'dag_definition'.
    """
    node_type = 'dag'

    def __init__(self, name, internal_name, config, execution_type, internal_branch_name=None):
        # pylint: disable=R0914,R0913
        super().__init__(name, internal_name, config, execution_type, internal_branch_name=internal_branch_name)
        self.sub_dag_file = self.config.get('dag_definition', None)
        self.is_composite = True

        if not self.sub_dag_file:
            raise Exception('A node type of dag requires a parameter dag_definition, please define it in the config')

        self.branch = self.get_sub_graph()

    @property
    def _internal_branch_name(self):
        """
        THe internal branch name in dot path convention

        Returns:
            [type]: [description]
        """
        return self.internal_name + '.' + defaults.DAG_BRANCH_NAME

    def get_sub_graph(self):
        """
        Create a sub-dag from the config['dag_definition']

        The graph object has an internal branch name, that is equal to the name of the step.
        And the sub-dag nodes follow an dot path naming convention

        Returns:
            Graph: A graph object
        """

        dag_config = utils.load_yaml(self.sub_dag_file)
        if 'dag' not in dag_config:
            raise Exception(f'No DAG found in {self.sub_dag_file}, please provide it in dag block')

        branch = create_graph(dag_config['dag'],
                              internal_branch_name=self._internal_branch_name)
        return branch

    def get_branch_by_name(self, branch_name: str):
        """
        Retrieve a branch by name.
        The name is expected to follow a dot path convention.

        Returns a Graph Object

        Args:
            branch_name (str): The name of the branch to retrieve

        Raises:
            Exception: If the branch_name is not 'dag'
        """
        if branch_name != self._internal_branch_name:
            raise Exception(f'Node of type {self.node_type} only allows a branch of name {defaults.DAG_BRANCH_NAME}')

        return self.branch

    def execute(self, executor, mock=False, map_variable: dict = None, **kwargs):
        """
        This method should never be called for a node of type dag

        Args:
            executor (BaseExecutor): The Executor class as defined by the config
            mock (bool, optional): If the operation is just a mock. Defaults to False.

        Raises:
            NotImplementedError: This method should never be called for a node of type Parallel
        """
        raise Exception('Node is of type composite, error in traversal rules')

    def execute_as_graph(self, executor, map_variable: dict = None, **kwargs):
        """
        This function does the actual execution of the branch of the dag node.

        From a design perspective, this function should not be called if the execution mode is 3rd party orchestrated.

        The modes that render the job specifications, do not need to interact with this node at all
        as they have their own internal mechanisms of handling sub dags.
        If they do not, you can find a way using as-is nodes as hack nodes.

        The actual logic is :
            * We just execute the branch as with any other composite nodes
            * The branch name is called 'dag'

        The execution of a dag, could result in
            * The dag being completely executed with a definite (fail, success) state in case of
                local or local-container execution
            * The dag being in a processing state with PROCESSING status in case of local-aws-batch

        Only fail state is considered failure during this phase of execution.

        Args:
            executor (Executor): The Executor as per the use config
            **kwargs: Optional kwargs passed around
        """
        step_success_bool = True
        waiting = False

        effective_branch_name = self.resolve_map_placeholders(self._internal_branch_name, map_variable=map_variable)
        effective_internal_name = self.resolve_map_placeholders(self.internal_name, map_variable=map_variable)

        branch_log = executor.run_log_store.create_branch_log(effective_branch_name)
        branch_log.status = defaults.PROCESSING
        executor.run_log_store.add_branch_log(branch_log, executor.run_id)

        executor.execute_graph(self.branch, map_variable=map_variable, **kwargs)

        branch_log = executor.run_log_store.get_branch_log(effective_branch_name, executor.run_id)
        if branch_log.status == defaults.FAIL:
            step_success_bool = False

        if branch_log.status == defaults.PROCESSING:
            waiting = True

        step_log = executor.run_log_store.get_step_log(effective_internal_name, executor.run_id)
        step_log.status = defaults.PROCESSING

        if step_success_bool:  #  If none failed and nothing is waiting
            if not waiting:
                step_log.status = defaults.SUCCESS
        else:
            step_log.status = defaults.FAIL

        executor.run_log_store.add_step_log(step_log, executor.run_id)


class AsISNode(BaseNode):
    """
    AsIs is a convenience design node.

    It always returns success in the attempt log and does nothing during interactive compute.

    The command given to execute is ignored but it does do the syncing of the catalog.
    This node is very akin to pass state in Step functions.

    This node type could be handy when designing the pipeline and stubbing functions

    But in render mode for job specification of a 3rd party orchestrator, this node comes handy.
    """
    node_type = 'as-is'

    def __init__(self, name, internal_name, config, execution_type, internal_branch_name=None):
        # pylint: disable=R0914,R0913
        super().__init__(name, internal_name, config, execution_type, internal_branch_name=internal_branch_name)

    def execute(self, executor, mock=False, map_variable: dict = None, **kwargs):
        """
        Do Nothing node.
        We just send an success attempt log back to the caller

        Args:
            executor ([type]): [description]
            mock (bool, optional): [description]. Defaults to False.
            map_variable (str, optional): [description]. Defaults to ''.

        Returns:
            [type]: [description]
        """
        attempt_log = executor.run_log_store.create_attempt_log()

        attempt_log.start_time = str(datetime.now())
        attempt_log.status = defaults.SUCCESS  # This is a dummy node and always will be success

        attempt_log.end_time = str(datetime.now())
        attempt_log.duration = utils.get_duration_between_datetime_strings(
            attempt_log.start_time, attempt_log.end_time)
        return attempt_log

    def execute_as_graph(self, executor, map_variable: dict = None, **kwargs):
        """
        Should not be implemented for a single node.

        Args:
            executor ([type]): [description]

        Raises:
            Exception: Not a composite node, always raises an exception
        """
        raise Exception('Node is not a composite node, invalid traversal rule')



NAME = 'magnus'

# CLI settings
LOG_LEVEL = 'WARNING'

# Interaction settings
TRACK_PREFIX = 'MAGNUS_TRACK_'
PARAMETER_PREFIX = 'MAGNUS_PRM_'

# STATUS progression
# For Branch, CREATED -> PROCESSING -> SUCCESS OR FAIL
# For a step, CREATED -> TRIGGERED ->  PROCESSING -> SUCCESS OR FAIL
CREATED = 'CREATED'
PROCESSING = 'PROCESSING'
SUCCESS = 'SUCCESS'
FAIL = 'FAIL'
TRIGGERED = 'TRIGGERED'

# Node and Command settings
COMMAND_TYPE = 'python'
NODE_SPEC_FILE = 'node_spec.yaml'
COMMAND_FRIENDLY_CHARACTER = '%'

# Default services
DEFAULT_EXECUTOR = {
    'type': 'local'
}
DEFAULT_RUN_LOG_STORE = {
    'type': 'buffered'
}
DEFAULT_CATALOG = {
    'type': 'file-system'
}
DEFAULT_SECRETS = {
    'type': 'do-nothing'
}

# Map state
MAP_PLACEHOLDER = 'map_variable_placeholder'

# Dag node
DAG_BRANCH_NAME = 'dag'

# RUN settings
RANDOM_RUN_ID_LEN = 6
MAX_TIME = 86400  # 1 day in seconds

# User extensions
USER_CONFIG_FILE = 'magnus-config.yaml'

# Executor settings
ENABLE_PARALLEL = False

# RUN log store settings
LOG_LOCATION_FOLDER = '.run_log_store'

# Dag node
DAG_BRANCH_NAME = 'dag'

# Data catalog settings
CATALOG_LOCATION_FOLDER = '.catalog'
COMPUTE_DATA_FOLDER = 'data'

# Secrets settings
DOTENV_FILE_LOCATION = '.env'

# AWS settings
AWS_REGION = 'eu-west-1'

# SEc



import logging
from typing import TYPE_CHECKING, Dict, List

from stevedore import driver

from magnus import defaults, exceptions

if TYPE_CHECKING:
    from magnus.nodes import BaseNode

logger = logging.getLogger(defaults.NAME)


class Graph:
    """
    A class representing a graph.

    The representation is similar to AWS step functions.
    We have nodes and traversal is based on start_at and on_failure definition of individual nodes of the graph
    """

    def __init__(self, start_at, description=None, max_time=86400, internal_branch_name=None):
        self.start_at = start_at
        self.description = description
        self.max_time = max_time
        self.internal_branch_name = internal_branch_name
        self.nodes = []

    def get_node_by_name(self, name: str) -> 'BaseNode':
        """
        Return the Node object by the name
        The name is always relative to the graph

        Args:
            name (str): Name of the node

        Raises:
            NodeNotFoundError: If the node of name is not found in the graph

        Returns:
            Node: The Node object by name
        """
        for node in self.nodes:
            if node.name == name:
                return node
        raise exceptions.NodeNotFoundError(name)

    def get_node_by_internal_name(self, internal_name: str) -> 'BaseNode':
        """
        Return the node by the internal name of the node.
        The internal name uses dot path convention.
        This method is only relative to the nodes of the graph and does not perform graph search of sub-graphs

        Args:
            internal_name (str): The internal name of the node, follows a dot path convention

        Raises:
            NodeNotFoundError: If the node of the internal name is not found in the graph

        Returns:
            Node: The Node object by the name
        """
        for node in self.nodes:
            if node.internal_name == internal_name:
                return node
        raise exceptions.NodeNotFoundError(internal_name)

    def __str__(self):  # pragma: no cover
        node_str = ', '.join([x.name for x in self.nodes])
        return f'Starts at: {self.start_at} and has a max run time of {self.max_time} and {node_str}'

    def add_node(self, node: 'BaseNode'):
        """
        Add a node to the nodes of the graph

        Args:
            node (object): The node to add
        """
        self.nodes.append(node)

    def validate(self):
        """
        Validate the graph to make sure,
        1). All the neighbors of nodes are present.
        2). Detection of cycles.
        3). Confirming that the start_at is actually a node present in graph
        4). Detection of one and only one success node.
        5). Detection of one and only one fail node.
        6). Missing nodes if referred by next or on_failure

        Raises:
            Exception: [description]
        """
        messages = []

        missing_nodes = self.missing_neighbors()
        if missing_nodes:
            message = 'The graph has references to nodes (next, on_failure), these nodes are missing from the DAG:\n'
            message += f'{", ".join(missing_nodes)}'
            logger.error(message)
            messages.append(message)

        if not missing_nodes:
            if not self.is_dag():
                message = 'The DAG is cyclic or does not reach an end state'
                logger.error(message)
                messages.append(message)

        if not self.is_start_node_present():
            message = 'The start node is not part of the graph'
            logger.error(message)
            messages.append(message)

        if not self.success_node_validation():
            message = 'There should be exactly one success node'
            logger.error(message)
            messages.append(message)

        if not self.fail_node_validation():
            message = 'There should be exactly one fail node'
            logger.error(message)
            messages.append(message)

        if messages:
            raise Exception(',  '.join(messages))

    def get_success_node(self) -> 'BaseNode':
        """
        Return the success node of the graph

        Raises:
            Exception: If no success node is present in the graph

        Returns:
            object: The success node
        """
        for node in self.nodes:
            if node.node_type == 'success':
                return node
        raise Exception('No success node defined')

    def get_fail_node(self) -> 'BaseNode':
        """
        Returns the fail node of the graph

        Raises:
            Exception: If no fail  node is present in the graph

        Returns:
            object: The fail node of the graph
        """
        for node in self.nodes:
            if node.node_type == 'fail':
                return node
        raise Exception('No fail node defined')

    def is_start_node_present(self) -> bool:
        """
        A check to ensure the start_at is part of the graph

        Returns:
            bool: True if start_at is one of the nodes, false otherwise
        """
        try:
            self.get_node_by_name(self.start_at)
            return True
        except exceptions.NodeNotFoundError:
            logger.exception('Could not find the node')
            return False

    def success_node_validation(self) -> bool:
        """
        Check to ensure there is one and only one success node in the graph

        Returns:
            bool: True if there is only one, false otherwise
        """
        node_count = 0
        for node in self.nodes:
            if node.node_type == 'success':
                node_count += 1
        if node_count == 1:
            return True
        return False

    def fail_node_validation(self) -> bool:
        """
        Check to make sure there is one and only one fail node in the graph

        Returns:
            bool: true if there is one and only one fail node, false otherwise
        """
        node_count = 0
        for node in self.nodes:
            if node.node_type == 'fail':
                node_count += 1
        if node_count == 1:
            return True
        return False

    def is_dag(self) -> bool:
        """
        Determines whether the graph is acyclic and directed

        Returns:
            bool: Returns True if it is directed and acyclic.
        """
        visited = {n.name: False for n in self.nodes}
        recstack = {n.name: False for n in self.nodes}

        for node in self.nodes:
            if not visited[node.name]:
                if self.is_cyclic_util(node, visited, recstack):
                    return False
        return True

    def is_cyclic_util(self, node: 'BaseNode', visited: Dict[str, bool], recstack: Dict[str, bool]) -> bool:
        """
        Recursive utility that determines if a node and neighbors has a cycle. Is used in is_dag method.

        Args:
            node (BaseNode): The node to check
            visited (dict): Dictionary storing which nodes have been checked
            recstack (dict): Stores what nodes have been visited recursively

        Returns:
            bool: True if cyclic.
        """
        visited[node.name] = True
        recstack[node.name] = True

        neighbors = node.get_neighbors()
        for neighbor in neighbors:
            neighbor_node = self.get_node_by_name(neighbor)
            if not visited[neighbor]:
                if self.is_cyclic_util(neighbor_node, visited, recstack):
                    return True
            elif recstack[neighbor]:
                return True

        recstack[node.name] = False
        return False

    def missing_neighbors(self) -> List['BaseNode']:
        """
        Iterates through nodes and gets their connecting neighbors and checks if they exist in the graph.

        Returns:
            list: List of the missing nodes. Empty list if all neighbors are in the graph.
        """
        missing_nodes = []
        for node in self.nodes:
            neighbors = node.get_neighbors()
            for neighbor in neighbors:
                try:
                    self.get_node_by_name(neighbor)
                except exceptions.NodeNotFoundError:
                    logger.exception('Could not find the node')
                    if neighbor not in missing_nodes:
                        missing_nodes.append(neighbor)
        return missing_nodes


def create_graph(dag_config: dict, internal_branch_name: str = None) -> Graph:
    # pylint: disable=R0914,R0913
    """
    Creates a dag object from the dag definition.

    Composite nodes like map, parallel, dag can have sub-branches which are internally graphs.
    Use internal_branch_name to fit the right dot path convention.

    Args:
        dag_config (dict): The dag defintion
        internal_branch_name ([type], optional): In case of sub-graph, the name of the node. Defaults to None.

    Raises:
        Exception: If the node or graph validation fails.

    Returns:
        Graph: The created graph object
    """
    # Conditional import to avoid circular import
    # pylint: disable=C0415
    from magnus.nodes import validate_node

    description = dag_config.get('description', None)
    max_time = dag_config.get('max_time', defaults.MAX_TIME)
    start_at = dag_config.get('start_at')  # Let the start_at be relative to the graph

    graph = Graph(start_at=start_at, description=description,
                  max_time=max_time, internal_branch_name=internal_branch_name)

    logger.info(
        f'Initialized a graph object that starts at {start_at} and runs for maximum of {max_time} secs')
    messages = []
    for step in dag_config.get('steps', []):
        step_config = dag_config['steps'][step]
        logger.info(f'Adding node {step} with :{step_config}')

        task_type = step_config.get('command_type', defaults.COMMAND_TYPE)
        command_config = step_config.get('command_config', {})

        logger.info(f"Trying to get a task of type {task_type}")
        try:
            task_mgr = driver.DriverManager(
                namespace="magnus.tasks.BaseTaskType",
                name=task_type,
                invoke_on_load=True,
                invoke_kwds={'command': step_config.get('command', None),
                             'config': command_config}
            )
        except Exception as _e:
            msg = (
                f"Could not find the task type {task_type}. Please ensure you have installed the extension that"
                " provides the task type. \nCore supports: python(default), python-lambda, shell, notebook"
            )
            raise Exception(msg) from _e

        internal_name = step
        if internal_branch_name:
            internal_name = internal_branch_name + '.' + step

        try:
            node_mgr = driver.DriverManager(
                namespace="magnus.nodes.BaseNode",
                name=step_config['type'],
                invoke_on_load=True,
                invoke_kwds={"name": step, "internal_name": internal_name,
                             "config": step_config, "execution_type": task_mgr.driver,
                             "internal_branch_name": internal_branch_name}
            )
        except Exception as _e:
            msg = (
                f"Could not find the node type {step_config['type']}. Please ensure you have installed "
                "the extension that provides the node type."
                "\nCore supports: task, success, fail, parallel, dag, map, as-is")
            raise Exception(msg) from _e

        messages.extend(validate_node(node_mgr.driver))
        graph.add_node(node_mgr.driver)

    if messages:
        raise Exception(', '.join(messages))

    graph.validate()

    return graph


def search_node_by_internal_name(dag: Graph, internal_name: str):
    """
    Given a DAG, search the node by internal name of the node.

    The node naming convention follows dot path naming convention

    Currently it is implemented to search only against the base dag.

    Args:
        dag (Graph): The graph to search the node
        internal_name (str): The internal name of the node.
    """
    # If the node is not part of any branches, then the base graph is where the node belongs
    dot_path = internal_name.split('.')
    if len(dot_path) == 1:
        return dag.get_node_by_internal_name(internal_name), dag

    # Any node internal name is: And is always going to be odd in number when split against .
    # Step.Branch.Step.Branch etc
    current_node = None
    current_branch = dag

    for i in range(len(dot_path)):
        if i % 2:
            # Its odd, so we are in brach name
            current_branch = current_node.get_branch_by_name('.'.join(dot_path[:i + 1]))  # type: ignore
            logger.debug(f'Finding step for {internal_name} in branch: {current_branch}')
        else:
            # Its even, so we are in Step, we start here!
            current_node = current_branch.get_node_by_internal_name('.'.join(dot_path[:i + 1]))
            logger.debug(f'Finding {internal_name} in node: {current_node}')

    logger.debug(f'current branch : {current_branch}, current step {current_node}')
    if current_branch and current_node:
        return current_node, current_branch

    raise exceptions.NodeNotFoundError(internal_name)


def search_branch_by_internal_name(dag: Graph, internal_name: str):
    """
    Given a DAG, search the branch by internal name of the branch.

    The branch naming convention follows dot path naming convention

    Currently it is implemented to search only against the base dag.

    Args:
        dag (Graph): The graph to search the node
        internal_name (str): The internal name of the branch.
    """
    # If the node is not part of any branches, then the base graph is where the node belongs
    dot_path = internal_name.split('.')
    if len(dot_path) == 1:
        return dag

    # Any branch internal name is: And is always going to be even in number when split against .
    # Step.Branch.Step.Branch
    current_node = None
    current_branch = dag

    for i in range(len(dot_path)):

        if i % 2:
            # Its odd, so we are in brach name
            current_branch = current_node.get_branch_by_name('.'.join(dot_path[:i + 1]))  # type: ignore
            logger.debug(f'Finding step for {internal_name} in branch: {current_branch}')

        else:
            # Its even, so we are in Step, we start here!
            current_node = current_branch.get_node_by_internal_name('.'.join(dot_path[:i + 1]))
            logger.debug(f'Finding {internal_name} in node: {current_node}')

    logger.debug(f'current branch : {current_branch}, current step {current_node}')
    if current_branch and current_node:
        return current_branch

    raise exceptions.BranchNotFoundError(internal_name)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['magnus']

package_data = \
{'': ['*']}

install_requires = \
['click',
 'click-plugins>=1.1.1,<2.0.0',
 'pydantic>=1.9.0,<2.0.0',
 'ruamel.yaml',
 'ruamel.yaml.clib',
 'stevedore>=3.5.0,<4.0.0',
 'yachalk']

extras_require = \
{'docker': ['docker'], 'notebook': ['papermill']}

entry_points = \
{'console_scripts': ['magnus = magnus.cli:cli'],
 'magnus.catalog.BaseCatalog': ['do-nothing = magnus.catalog:DoNothingCatalog',
                                'file-system = '
                                'magnus.catalog:FileSystemCatalog'],
 'magnus.datastore.BaseRunLogStore': ['buffered = '
                                      'magnus.datastore:BufferRunLogstore',
                                      'file-system = '
                                      'magnus.datastore:FileSystemRunLogstore'],
 'magnus.executor.BaseExecutor': ['demo-renderer = '
                                  'magnus.executor:DemoRenderer',
                                  'local = magnus.executor:LocalExecutor',
                                  'local-container = '
                                  'magnus.executor:LocalContainerExecutor'],
 'magnus.integration.BaseIntegration': ['demo-renderer-run_log_store-buffered '
                                        '= '
                                        'magnus.integration:DemoRenderBufferedRunLogStore',
                                        'local-catalog-do-nothing = '
                                        'magnus.integration:LocalDoNothingCatalog',
                                        'local-container-catalog-do-nothing = '
                                        'magnus.integration:LocalContainerDoNothingCatalog',
                                        'local-container-catalog-file-system = '
                                        'magnus.integration:LocalContainerComputeFileSystemCatalog',
                                        'local-container-run_log_store-buffered '
                                        '= '
                                        'magnus.integration:LocalContainerComputeBufferedRunLogStore',
                                        'local-container-run_log_store-file-system '
                                        '= '
                                        'magnus.integration:LocalContainerComputeFileSystemRunLogstore',
                                        'local-container-secrets-dotenv = '
                                        'magnus.integration:LocalContainerComputeDotEnvSecrets',
                                        'local-run_log_store-buffered = '
                                        'magnus.integration:LocalComputeBufferedRunLogStore',
                                        'local-run_log_store-file-system = '
                                        'magnus.integration:LocalComputeFileSystemRunLogStore'],
 'magnus.nodes.BaseNode': ['as-is = magnus.nodes:AsISNode',
                           'dag = magnus.nodes:DagNode',
                           'fail = magnus.nodes:FailNode',
                           'map = magnus.nodes:MapNode',
                           'parallel = magnus.nodes:ParallelNode',
                           'success = magnus.nodes:SuccessNode',
                           'task = magnus.nodes:TaskNode'],
 'magnus.secrets.BaseSecrets': ['do-nothing = '
                                'magnus.secrets:DoNothingSecretManager',
                                'dotenv = magnus.secrets:DotEnvSecrets'],
 'magnus.tasks.BaseTaskType': ['notebook = magnus.tasks:NotebookTaskType',
                               'python = magnus.tasks:PythonTaskType',
                               'python-lambda = '
                               'magnus.tasks:PythonLambdaTaskType',
                               'shell = magnus.tasks:ShellTaskType']}

setup_kwargs = {
    'name': 'magnus',
    'version': '0.2.2',
    'description': 'A Compute agnostic pipelining software',
    'long_description': '# Hello from magnus\n\nMagnus is a data science pipeline definition and execution tool. It provides a way to:\n\n- Define a pipeline steps and the flow.\n- Run the pipeline in any environment, local is default.\n- Store the run metadata and data catalogs and re-run in case of failures.\n\nOnce the pipeline is proven to be correct and functional in any environment, there is zero code change\nrequired to deploy it elsewhere. The behavior of the runs are identical in all environments. Magnus\nis not a queuing or scheduling engine, but delegates that responsibility to chosen deployment patterns.\n\n### Short Summary\n\nMagnus provides four capabilities for data teams:\n\n- **Compute execution plan**: A DAG representation of work that you want to get done. Individual nodes of the DAG\ncould be simple python or shell tasks or complex deeply nested parallel branches or embedded DAGs themselves.\n\n- **Run log store**: A place to store run logs for reporting or re-running older runs. Along with capturing the\nstatus of execution,  the run logs also capture code identifiers (commits, docker image digests etc), data hashes and\nconfiguration settings for reproducibility and audit.\n\n- **Data Catalogs**: A way to pass data between nodes of the graph during execution and also serves the purpose of\nversioning the data used by a particular run.\n\n- **Secrets**: A framework to provide secrets/credentials at run time to the nodes of the graph.\n\n### Design decisions:\n\n- **Easy to extend**: All the four capabilities are just definitions and can be implemented in many flavors.\n\n    - **Compute execution plan**: You can choose to run the DAG on your local computer, in containers of local computer\n    or off load the work to cloud providers or translate the DAG to AWS step functions or Argo workflows.\n\n    - **Run log Store**: The actual implementation of storing the run logs could be in-memory, file system, S3,\n    database etc.\n\n    - **Data Catalogs**: The data files generated as part of a run could be stored on file-systems, S3 or could be\n    extended to fit your needs.\n\n    - **Secrets**: The secrets needed for your code to work could be in dotenv, AWS or extended to fit your needs.\n\n- **Pipeline as contract**: Once a DAG is defined and proven to work in local or some environment, there is absolutely\nno code change needed to deploy it to other environments. This enables the data teams to prove the correctness of\nthe dag in dev environments while infrastructure teams to find the suitable way to deploy it.\n\n- **Reproducibility**: Run log store and data catalogs hold the version, code commits, data files used for a run\nmaking it easy to re-run an older run or debug a failed run. Debug environment need not be the same as\noriginal environment.\n\n- **Easy switch**: Your infrastructure landscape changes over time. With magnus, you can switch infrastructure\nby just changing a config and not code.\n\n\nMagnus does not aim to replace existing and well constructed orchestrators like AWS Step functions or\n[argo](https://argoproj.github.io/workflows/) but complements them in a unified, simple and intuitive way.\n\n## Documentation\n\n[More details about the project and how to use it available here](https://astrazeneca.github.io/magnus-core/).\n\n## Installation\n\n### pip\n\nmagnus is a python package and should be installed as any other.\n\n```shell\npip install magnus\n```\n\n# Example Run\n\nTo give you a flavour of how magnus works, lets create a simple pipeline.\n\nCopy the contents of this yaml into getting-started.yaml.\n\n---\n!!! Note\n\n   The below execution would create a folder called \'data\' in the current working directory.\n   The command as given should work in linux/macOS but for windows, please change accordingly.\n\n---\n\n``` yaml\ndag:\n  description: Getting started\n  start_at: step parameters\n  steps:\n    step parameters:\n      type: task\n      command_type: python-lambda\n      command: "lambda x: {\'x\': int(x) + 1}"\n      next: step shell\n    step shell:\n      type: task\n      command_type: shell\n      command: mkdir data ; env >> data/data.txt # For Linux/macOS\n      next: success\n      catalog:\n        put:\n          - "*"\n    success:\n      type: success\n    fail:\n      type: fail\n```\n\nSince the pipeline expects a parameter ```x```, lets provide that using ```parameters.yaml```\n\n```yaml\nx: 3\n```\n\nAnd let\'s run the pipeline using:\n``` shell\n magnus execute --file getting-started.yaml --parameters-file parameters.yaml\n```\n\nYou should see a list of warnings but your terminal output should look something similar to this:\n\n``` json\n{\n    "run_id": "20220118114608",\n    "dag_hash": "ce0676d63e99c34848484f2df1744bab8d45e33a",\n    "use_cached": false,\n    "tag": null,\n    "original_run_id": "",\n    "status": "SUCCESS",\n    "steps": {\n        "step parameters": {\n            "name": "step parameters",\n            "internal_name": "step parameters",\n            "status": "SUCCESS",\n            "step_type": "task",\n            "message": "",\n            "mock": false,\n            "code_identities": [\n                {\n                    "code_identifier": "c5d2f4aa8dd354740d1b2f94b6ee5c904da5e63c",\n                    "code_identifier_type": "git",\n                    "code_identifier_dependable": false,\n                    "code_identifier_url": "<INTENTIONALLY REMOVED>",\n                    "code_identifier_message": "<INTENTIONALLY REMOVED>"\n                }\n            ],\n            "attempts": [\n                {\n                    "attempt_number": 0,\n                    "start_time": "2022-01-18 11:46:08.530138",\n                    "end_time": "2022-01-18 11:46:08.530561",\n                    "duration": "0:00:00.000423",\n                    "status": "SUCCESS",\n                    "message": ""\n                }\n            ],\n            "user_defined_metrics": {},\n            "branches": {},\n            "data_catalog": []\n        },\n        "step shell": {\n            "name": "step shell",\n            "internal_name": "step shell",\n            "status": "SUCCESS",\n            "step_type": "task",\n            "message": "",\n            "mock": false,\n            "code_identities": [\n                {\n                    "code_identifier": "c5d2f4aa8dd354740d1b2f94b6ee5c904da5e63c",\n                    "code_identifier_type": "git",\n                    "code_identifier_dependable": false,\n                    "code_identifier_url": "<INTENTIONALLY REMOVED>",\n                    "code_identifier_message": "<INTENTIONALLY REMOVED>"\n                }\n            ],\n            "attempts": [\n                {\n                    "attempt_number": 0,\n                    "start_time": "2022-01-18 11:46:08.576522",\n                    "end_time": "2022-01-18 11:46:08.588158",\n                    "duration": "0:00:00.011636",\n                    "status": "SUCCESS",\n                    "message": ""\n                }\n            ],\n            "user_defined_metrics": {},\n            "branches": {},\n            "data_catalog": [\n                {\n                    "name": "data.txt",\n                    "data_hash": "8f25ba24e56f182c5125b9ede73cab6c16bf193e3ad36b75ba5145ff1b5db583",\n                    "catalog_relative_path": "20220118114608/data.txt",\n                    "catalog_handler_location": ".catalog",\n                    "stage": "put"\n                }\n            ]\n        },\n        "success": {\n            "name": "success",\n            "internal_name": "success",\n            "status": "SUCCESS",\n            "step_type": "success",\n            "message": "",\n            "mock": false,\n            "code_identities": [\n                {\n                    "code_identifier": "c5d2f4aa8dd354740d1b2f94b6ee5c904da5e63c",\n                    "code_identifier_type": "git",\n                    "code_identifier_dependable": false,\n                    "code_identifier_url": "<INTENTIONALLY REMOVED>",\n                    "code_identifier_message": "<INTENTIONALLY REMOVED>"\n                }\n            ],\n            "attempts": [\n                {\n                    "attempt_number": 0,\n                    "start_time": "2022-01-18 11:46:08.639563",\n                    "end_time": "2022-01-18 11:46:08.639680",\n                    "duration": "0:00:00.000117",\n                    "status": "SUCCESS",\n                    "message": ""\n                }\n            ],\n            "user_defined_metrics": {},\n            "branches": {},\n            "data_catalog": []\n        }\n    },\n    "parameters": {\n        "x": 4\n    },\n    "run_config": {\n        "executor": {\n            "type": "local",\n            "config": {}\n        },\n        "run_log_store": {\n            "type": "buffered",\n            "config": {}\n        },\n        "catalog": {\n            "type": "file-system",\n            "config": {}\n        },\n        "secrets": {\n            "type": "do-nothing",\n            "config": {}\n        }\n    }\n}\n```\n\nYou should see that ```data``` folder being created with a file called ```data.txt``` in it.\nThis is according to the command in ```step shell```.\n\nYou should also see a folder ```.catalog``` being created with a single folder corresponding to the run_id of this run.\n\nTo understand more about the input and output, please head over to the\n[documentation](https://project-magnus.github.io/magnus-core/).\n',
    'author': 'Vijay Vammi',
    'author_email': 'vijay.vammi@astrazeneca.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/AstraZeneca/magnus-core',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)



import json
import logging
import os
from pathlib import Path
from typing import Union

from magnus import defaults, exceptions, utils

logger = logging.getLogger(defaults.NAME)


def track_this(**kwargs):
    """
    Set up the keyword args as environment variables for tracking purposes as
    part pf the run.

    For every key-value pair found in kwargs, we set up an environmental variable of
    MAGNUS_TRACK_key = json.dumps(value)

    Args:
        kwargs (dict): The dictionary of key value pairs to track.
    """
    for key, value in kwargs.items():
        logger.info(f'Tracking {key} with value: {value}')
        os.environ[defaults.TRACK_PREFIX + key] = json.dumps(value)


def store_parameter(**kwargs: dict):
    """
    Set up the keyword args as environment variables for parameters tracking
    purposes as part pf the run.

    For every key-value pair found in kwargs, we set up an environmental variable of
    MAGNUS_PRM_key = json.dumps(value)
    """
    for key, value in kwargs.items():
        logger.info(f'Storing parameter {key} with value: {value}')
        os.environ[defaults.PARAMETER_PREFIX + key] = json.dumps(value)


def get_parameter(key=None) -> Union[str, dict]:
    """
    Get the parameter set as part of the user convenience function.

    We do not remove the parameter from the environment in this phase as
    as the function execution has not been completed.

    Returns all the parameters, if no key was provided.

    Args:
        key (str, optional): The parameter key to retrieve. Defaults to None.

    Raises:
        Exception: If the menionted key was not part of the paramters

    Returns:
        Union[str, dict]: A single value of str or a dictionary if no key was specified
    """
    parameters = utils.get_user_set_parameters(remove=False)
    if not key:
        return parameters
    if key not in parameters:
        raise Exception(f'Parameter {key} is not set before')
    return parameters[key]


def get_secret(secret_name: str = None) -> str:
    """
    Get a secret by the name from the secrets manager

    Args:
        secret_name (str): The name of the secret to get. Defaults to None.

    Returns:
        str: The secret from the secrets manager, if exists. If the requested secret was None, we return all.
        Otherwise, raises exception.

    Raises:
        exceptions.SecretNotFoundError: Secret not found in the secrets manager.
    """
    from magnus.pipeline import \
        global_executor  # pylint: disable=import-outside-toplevel
    secrets_handler = global_executor.secrets_handler  # type: ignore

    try:
        return secrets_handler.get(name=secret_name)
    except exceptions.SecretNotFoundError:
        logger.exception(f'No secret by the name {secret_name} found in the store')
        raise


def get_from_catalog(name: str, destination_folder: str = None):
    """
    A convenience interaction function to get file from the catalog and place it in the destination folder

    Note: We do not perform any kind of serialization/deserialization in this way.
    Args:
        name (str): The name of the file to get from the catalog
        destination_folder (None): The place to put the file. defaults to compute data folder

    """
    from magnus.pipeline import \
        global_executor  # pylint: disable=import-outside-toplevel

    if not destination_folder:
        destination_folder = global_executor.catalog_handler.compute_data_folder  # type: ignore

    global_executor.catalog_handler.get(name, run_id=global_executor.run_id,  # type: ignore
                                        compute_data_folder=destination_folder)


def put_in_catalog(filepath: str):
    """
    A convenience interaction function to put the file in the catalog.

    Note: We do not perform any kind of serialization/deserialization in this way.

    Args:
        filepath (str): The path of the file to put in the catalog
    """
    from magnus.pipeline import \
        global_executor  # pylint: disable=import-outside-toplevel

    file_path = Path(filepath)

    global_executor.catalog_handler.put(file_path.name, run_id=global_executor.run_id,  # type: ignore
                                        compute_data_folder=file_path.parent)
